/*
Copyright 2009;  D. Liebman,

This file is part of gaptsearch.

gaptsearch is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

gaptsearch is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gaptsearch.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string.h>
#include <gtk/gtk.h>
#include "gaptsearch-glade.h"
#include "gaptsearch.h"

/* -------------function implementations-------------------- */

void
insert_text_special (   gchar * standard_output, 
                        Data * data )
{
    GError **gerror;
    GtkTextBuffer *buffer;
    gchar * command_line;
    gchar * output;
    gchar * command;
    GString * list;
    gint startpos, endpos;
    GtkTreeSelection *sel;
    GtkTreeModel *model;
    gchar *selected, *installed;
    GtkTreeIter selected_row;
    
    if(data->special_tab == CACHED_FILES )
    {

        command_line = "ls -1 /var/cache/apt/archives ";
        g_spawn_command_line_sync (  command_line,
                                    &output,
                                    NULL,
                                    NULL,
                                    gerror );
         
        /* process 'output' */
        list = g_string_new ( output);
        get_text_special_positions( list->str, "lock", &startpos, &endpos);
        g_string_erase ( list, startpos, endpos - startpos );
        get_text_special_positions( list->str, "partial", &startpos, &endpos);
        g_string_erase ( list, startpos, endpos - startpos );
    
        buffer = gtk_text_view_get_buffer( data->text_special );
        gtk_text_buffer_set_text (      buffer,
                                        list->str,
                                        -1 );
        g_string_free(list, TRUE);
        
        /* change label on text_special_label */
        gtk_label_set_markup (    data->text_special_label, 
                                "<b>Special Tab: All Cached Files</b>\n(<i>/var/cache/apt/archives</i>)" );
        
        
    }
    else if (data->special_tab == PACKAGE_INFO )
    {
        buffer = gtk_text_view_get_buffer( data->text_special );
        gtk_text_buffer_set_text (      buffer,
                                        standard_output,
                                        -1 );
                                        
        /* change label on text_special_label */
        gtk_label_set_markup (    data->text_special_label, 
                                "<b>Special Tab: Package Info</b>" );
    }                                   
    else if (data->special_tab == INSTALLED_FILES )
    {
    
        sel = gtk_tree_view_get_selection ( GTK_TREE_VIEW (data->tree));
    
        /* check what most recent selection was */
        if( gtk_tree_selection_get_selected(sel, &model, &selected_row) ) 
        {
            /* find which selection was pressed */
            gtk_tree_model_get ( GTK_TREE_MODEL( model ),
                                &selected_row,
                                0,
                                &installed,
                                1,
                                &selected,
                                -1);                                      
            
    
            command = "dpkg-query -L ";
            command_line =  g_strconcat( command , selected, NULL);
            g_spawn_command_line_sync (  command_line,
                                        &output,
                                        NULL,
                                        NULL,
                                        gerror );
             
            
            if (strlen(output) == 0)
                output = "Packages must be installed for files to be listed";
            /*    
            if (strlen(standard_output) != 0)
                output = standard_output;
            */
            buffer = gtk_text_view_get_buffer( data->text_special );
            gtk_text_buffer_set_text (      buffer,
                                            output,
                                            -1 );
                                            
            /* change label on text_special_label */
            gtk_label_set_markup (    data->text_special_label, 
                                    "<b>Special Tab: Individual Files Installed By This Package</b>" );
        }
    }



}

void
on_cached_files_toggled (GtkToggleAction * widget,
                            Data * data ) 
{
    
    gboolean test = gtk_toggle_action_get_active (widget);
    
    if (test == TRUE) {
        printf("cached files\n");
        data->special_tab = CACHED_FILES ;
        insert_text_special("", data);
    }
   
}

void
on_installed_files_toggled (GtkToggleAction * widget,
                            Data * data ) 
{
    
    gboolean test = gtk_toggle_action_get_active (widget);
    
    if (test == TRUE) {
        printf("installed files\n");
        data->special_tab = INSTALLED_FILES ;
        insert_text_special("", data);
    }
   
}

void
on_package_info_toggled (GtkToggleAction * widget,
                            Data * data ) 
{
    
    gboolean test = gtk_toggle_action_get_active (widget);
    
    if (test == TRUE) {
        printf("package info\n");
        data->special_tab = PACKAGE_INFO ;
        insert_description(data);
        //insert_text_special("", data);
    }
   
}

void
show_aptitude_toggled_cb(   GtkToggleAction * widget, 
                            Data *data )
{
    
    gboolean test = gtk_toggle_action_get_active (widget);
    
    if (test == TRUE) {
        //printf("show aptitude\n");
        data->show_command = USE_APTITUDE;
    }
    else if (test == FALSE) {
        //printf("show apt-cache\n");
        data->show_command = USE_APTCACHE;
    }

}

void
search_aptitude_toggled_cb( GtkToggleAction * widget, 
                            Data *data )
{

    gboolean test = gtk_toggle_action_get_active (widget);
    
    if (test == TRUE) {
        //printf("search aptitude\n");
        data->search_command = USE_APTITUDE;
    }
    else if (test == FALSE) {
        //printf("search apt-cache\n");
        data->search_command = USE_APTCACHE;
    }
}

static void
get_text_special_positions (    gchar * input,
                        gchar * delimiter,
                        gint  * a, 
                        gint  * b )
{    
   
    gint startpos;
    gint endpos;
    
    /* parse standard_output */
    gint x = 0;
    gint y = 0;
    
    GString * test = g_string_new("");
    GString * marker = g_string_new(delimiter);
    
    startpos = 0;
    while ('\0' !=  input[x]  ) 
    {
        
        g_string_append_c (test, input[x]);
        
        // find newline 
        if( x != 0 && '\n' == input[x] )
        {       
                test = g_string_new("");
                y = x;  
                     
        }
        
        if ( 0 == g_strcmp0( test->str , marker->str)) 
        {
                startpos = y ;
                x = strlen( input ) -1;
        }
        
        x ++;
        
    }
    
    endpos = startpos;
    x = startpos ;
    y = x;
    while ('\0' != input[x]  && endpos == startpos ) 
    {
        if( x != startpos && '\n' == input[x]  && ' ' != input[x+1] )
        {       
                endpos = x;           
        }
        x ++;
    }
    
    g_string_free(test, TRUE);
    g_string_free(marker, TRUE);
    
    *a = startpos;
    *b = endpos;
}


static void
get_text_positions (    gchar * input,
                        gchar * delimiter,
                        gint  * a, 
                        gint  * b )
{    
   
    gint startpos;
    gint endpos;
    
    /* parse standard_output */
    gint x = 0;
    gint y = 0;
    
    GString * test = g_string_new("");
    GString * marker = g_string_new(delimiter);
    
    startpos = 0;
    while ('\0' !=  input[x]  ) 
    {
        
        g_string_append_c (test, input[x]);
        
        // find newline 
        if( x != 0 && '\n' == input[x] )
        {       
                test = g_string_new("");  
                     
        }
            
        if ( 0 == g_strcmp0( test->str , marker->str)) 
        {
                startpos = x + 1;
        }
        
        x ++;
    }
    
    endpos = startpos;
    x = startpos ;
    y = x;
    while ('\0' != input[x]  && endpos == startpos ) 
    {
        if( x != startpos && '\n' == input[x]  && ' ' != input[x+1])
        {       
                endpos = x;           
        }
        x ++;
    }
    
    g_string_free(test, TRUE);
    g_string_free(marker, TRUE);
    
    *a = startpos;
    *b = endpos;
}

static void search_click(    GtkWidget *widget,
                                Data *data )
{

    /* Tree view vars */
    GtkListStore  *store;
    GtkTreeIter    tree_iter;
    
    const gchar *entry_text;
    
    /* g_spawn_command_line_sync variables*/
    gchar *command_line;
    gchar * standard_output;
    GError **gerror;
    gchar * apt_part;
    
    /* tree selection */
    GtkTreeSelection *sel;
   
    sel = gtk_tree_view_get_selection ( GTK_TREE_VIEW (data->tree));
    gtk_tree_selection_unselect_all(sel);
 
    entry_text = gtk_entry_get_text (GTK_ENTRY( data->entry) );
    
    /* select right commandline tool */
    apt_part = "/usr/bin/apt-cache search ";
    if ( data->search_command == USE_APTCACHE ) 
    {
        apt_part = "/usr/bin/apt-cache search ";
    }
    else if ( data->search_command == USE_APTITUDE )
    {
        apt_part = "/usr/bin/aptitude --disable-columns search ";
    }
    
    /* attempt at running apt-cache */
    command_line =  g_strconcat( apt_part , entry_text, NULL);
    g_spawn_command_line_sync (  command_line,
                        &standard_output,
                        NULL,
                        NULL,
                        gerror );
     
    g_free(command_line);
        
    store = GTK_LIST_STORE( gtk_tree_view_get_model( data->tree ) );
    gtk_list_store_clear(store);

    
    /* parse standard_output */
    gint x = 0;
    gint y = 0;
    gchar * backgroundcolor;
    gchar * icon;
    while ('\0' != standard_output[x] && '\n' != standard_output[x]) {
    
        GString * package;
        GString * description;
        GString * installed;    
        
        package = g_string_new("");
        description = g_string_new("");
        installed = g_string_new("");
    
        /* weather or not package was installed */
        while ( ' ' != standard_output[x] &&
                '\0' != standard_output[x] &&
                data->search_command == USE_APTITUDE ) 
        {
            
            g_string_append_c ( installed,
                                (gchar) standard_output[x]);
            x++;
        }
        /* advance past unused characters */   
        while ( ' ' == standard_output[x] ) 
                x++;
        
        if (    ' ' != standard_output[x] &&  
                '\0' != standard_output[x] &&
                '\n' != standard_output[x] &&
                ' ' == standard_output[x+1] ) 
                x= x + 2;
        
        /* find package name */
        while (  ' ' != standard_output[x] &&  
                '\0' != standard_output[x] &&
                '\n' != standard_output[x] ) 
        {
            
            g_string_append_c ( package,
                                (gchar) standard_output[x]);
                                
            x++;
        }
        
        /* advance past unused characters */        
        while ( ' ' == standard_output[x] ) 
                x++;

        while ( '\n' != standard_output[x] &&
                '\0' != standard_output[x] ) 
        {
            g_string_append_c (description,
                                (gchar) standard_output[x]);
            x++;
        }
        
        x++;
        
        y = y + 1;
        if (y % 2 == 0)
            backgroundcolor = "gray90";// very light gray
        else
            backgroundcolor = "white";
        
        if ( data->search_command == USE_APTITUDE ) 
        {    
            if ( installed->str[0] == 'i' )
                icon = GTK_STOCK_APPLY;
            else
                icon = GTK_STOCK_CANCEL;         
        }   
        gtk_list_store_append( store, &tree_iter );
        gtk_list_store_set( store, &tree_iter, 
                        0, installed->str, 
                        1, package->str, 
                        2, description->str,
                        3, backgroundcolor,
                        4, icon,
                        -1 );
        
        g_string_free(package, TRUE);
        g_string_free(description, TRUE);
        g_string_free(installed, TRUE);
    }
   
    /* clear text areas after search */
    GtkTextBuffer *buffer;
    buffer = gtk_text_view_get_buffer( data->text_all );
    gtk_text_buffer_set_text( buffer, "Select a package...", -1);
    
    buffer = gtk_text_view_get_buffer( data->text_description );
    gtk_text_buffer_set_text( buffer, "Select a package...", -1);
    
    /* text_special area */
    if (data->special_tab == INSTALLED_FILES ) 
    {
        buffer = gtk_text_view_get_buffer( data->text_special );
        gtk_text_buffer_set_text( buffer, "Select a package...", -1);
    }
    else
    {
        insert_text_special( "Select a package...", data);
    }
    
    /* clear statusbar after search */
    if ( data->status_id != 0 ) 
    {
        gtk_statusbar_remove (  GTK_STATUSBAR ( data->statusbar ),
                                data->context_id,
                                data->status_id );
    }
    
    /* change window title */
    gtk_window_set_title (  GTK_WINDOW(data->gladewindow),
                            "gaptsearch" );  

}

void
insert_status ( GtkWidget * widget,
                Data * data )
{
    GtkTreeIter selected_row;
    GtkTreeModel *model;
    gchar *selected, *installed;
    GString *title;
    GtkTreeSelection *sel;
   
    sel = gtk_tree_view_get_selection ( GTK_TREE_VIEW ( widget ));
    
    if( gtk_tree_selection_get_selected(sel, &model, &selected_row) ) 
    {
        /* find which selection was pressed */
        gtk_tree_model_get ( GTK_TREE_MODEL( model ),
                            &selected_row,
                            0,
                            &installed,
                            1,
                            &selected,
                            -1);
                            
        /* print info about installed status to status bar */
        if ( installed[0] == 'i' )
            installed = "   Installed (\"i\")";
            
        if ( installed[0] == 'p' )
            installed = "   No Files On System (\"p\")";
            
        if ( installed[0] == 'v' )
            installed = "   Virtual (\"v\")";
            
        if ( installed[0] == 'c' )
            installed = "   Config Files Remain (\"c\")";
        
        /* remove old message if there was one */
        if ( data->status_id != 0 ) 
        {
            gtk_statusbar_remove (  GTK_STATUSBAR ( data->statusbar ),
                                    data->context_id,
                                    data->status_id );
        }
        
        data->status_id = gtk_statusbar_push (  GTK_STATUSBAR ( data->statusbar ),
                                                data->context_id,
                                                installed);
        
        /* change window title */
        title = g_string_new ( "gaptsearch: " );
        g_string_append( title, selected );
        gtk_window_set_title (  GTK_WINDOW(data->gladewindow),
                                title->str );
        g_string_free (title, TRUE);  
        
    }   
    
}

/* insert description into text tabs */
static void 
insert_description( Data *data )
{
    GtkTextBuffer *buffer;
    GtkTreeIter selected_row;
    GtkTreeModel *model;
    gchar *selected, *installed;
    GtkTreeSelection *sel;
    GString *combined;
   
    sel = gtk_tree_view_get_selection ( GTK_TREE_VIEW (data->tree));
    
    /* check what most recent selection was */
    if( gtk_tree_selection_get_selected(sel, &model, &selected_row) ) 
    {
        /* find which selection was pressed */
        gtk_tree_model_get ( GTK_TREE_MODEL( model ),
                            &selected_row,
                            0,
                            &installed,
                            1,
                            &selected,
                            -1);                                      
        
    /* g_spawn_command_line_sync variables*/
        gchar *command_line;
        gchar * standard_output;
        GError **gerror;
        gint startpos;
        gint endpos;
        gchar * apt_part;
        
        /* select right commandline tool */
        apt_part = "/usr/bin/apt-cache show ";
        if ( data->show_command == USE_APTCACHE ) 
        {
                apt_part = "/usr/bin/apt-cache show ";
        }
        else if ( data->show_command == USE_APTITUDE )
        {
                apt_part = "/usr/bin/aptitude show ";
        }
        
    
    
        /* attempt at running apt-cache */
        command_line =  g_strconcat( apt_part , selected, NULL);
        g_spawn_command_line_sync (  command_line,
                                    &standard_output,
                                    NULL,
                                    NULL,
                                    gerror );
        
        /* change label on text_description_label */
        combined = g_string_new ("<b>Package Description</b>\n");
        g_string_append ( combined, command_line );
        gtk_label_set_markup (     data->text_description_label, 
                                combined->str );
        
        /* change label on text_all_label */
        g_string_assign ( combined, "<b>All Package Info</b>\n");
        g_string_append ( combined, command_line );
        gtk_label_set_markup (     data->text_all_label, 
                                    combined->str );
        
        g_string_free ( combined, TRUE);
        
        g_free(command_line);
        g_free(selected);
        
        /* Add this text to text_all view */
        buffer = gtk_text_view_get_buffer( data->text_all );
        gtk_text_buffer_set_text( buffer, standard_output, -1);
        
        /* find text for text_description view */
        get_text_positions (standard_output , "Description:" , &startpos, &endpos);
        
        buffer = gtk_text_view_get_buffer( data->text_description );
        gchar * description;
        
        description = g_strndup( &standard_output[startpos], endpos - startpos);
        
        /* remove unsightly seperator character from description paragraphs */
        gint x = 0;
        while ( description[x] != '\0' )
        {
            if( description[x+1] == '\n' && description[x] == '.' && x != 0 &&
                ( description[x-1] == ' '  ||  description[x-1] == '\n' ) ) 
            {
                description[x] = ' ';
            }
            x ++;
        }
        
        /* put description in text_description buffer */
        gtk_text_buffer_set_text (      buffer, 
                                        description, 
                                        -1);
        
        /* find text for text_special view */
        if (data->special_tab == PACKAGE_INFO)
        {
            insert_text_special (   standard_output,
                                    data );
        }
        else 
        {
            insert_text_special ( "", data);
        }

        g_free(standard_output);
        
    }

}
   
/*   main   */
   
int 
main(   int   argc,
        char *argv[] )
{

    GtkWidget *text_all;
    GtkWidget *text_special;
    GtkWidget *text_description;
    GtkWidget *entry, *button;
    GtkWidget *text_all_label;
    GtkWidget *text_description_label;
    GtkWidget *text_special_label;

    /* Data  */    
    Data data;
    
    /* glade oriented variables */
    GtkBuilder  *builder;
    GtkTreeView *treeview;
    GtkCellRenderer *cell, *pix;
    GtkTreeViewColumn *column0, *column1, *column2, *column3;
    GtkListStore *store;
    GtkTreeSelection *sel;
    GtkStatusbar *statusbar;

    gtk_init (&argc, &argv);
   
    data.search_command = USE_APTITUDE;
    data.show_command = USE_APTCACHE;
    data.special_tab = CACHED_FILES;
    
    builder = gtk_builder_new ();
    gtk_builder_add_from_string (	builder,glade_ui , 
									strlen(glade_ui) * sizeof(gchar),
									NULL);
    
    data.gladewindow = GTK_WIDGET ( gtk_builder_get_object (builder, "window") );
    gtk_builder_connect_signals (builder,  &data);
    
    /* setup glade list */
    store = gtk_list_store_new (5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
    treeview = GTK_TREE_VIEW( gtk_builder_get_object (builder, "treeview" ) );
    gtk_tree_view_set_model (treeview, GTK_TREE_MODEL(store));
    
    cell = gtk_cell_renderer_text_new ();
    pix = gtk_cell_renderer_pixbuf_new();
    
    column0 = gtk_tree_view_column_new_with_attributes ("",
                                                        pix,
                                                        "stock-id", 4,
                                                        "cell-background", 3, 
                                                        NULL);
    
    column1 = gtk_tree_view_column_new_with_attributes ("I",
                                                       cell,
                                                       "text", 0,
                                                       "background",3,
                                                       NULL);

    column2 = gtk_tree_view_column_new_with_attributes ("Packages",
                                                       cell,
                                                       "text", 1,
                                                       "background",3,
                                                       NULL);
    column3 = gtk_tree_view_column_new_with_attributes ("Description",
                                                       cell,
                                                       "text", 2,
                                                       "background",3,
                                                       NULL);
  
    gtk_tree_view_append_column (   GTK_TREE_VIEW (treeview),
	  		                        column0);
    gtk_tree_view_append_column (   GTK_TREE_VIEW (treeview),
	  		                        column1);
    gtk_tree_view_append_column (   GTK_TREE_VIEW (treeview),
	  		                        column2);
    gtk_tree_view_append_column (   GTK_TREE_VIEW (treeview),
	  		                        column3);
    
    
    sel = gtk_tree_view_get_selection ( GTK_TREE_VIEW (treeview));
    gtk_tree_selection_set_mode(sel, GTK_SELECTION_SINGLE);
    
    g_signal_connect_swapped( sel, "changed", 
                                G_CALLBACK(insert_description), 
                                &data);
    
    /* connect statusbar */
    statusbar = GTK_STATUSBAR( gtk_builder_get_object (builder, "statusbar" ) );
    data.context_id = gtk_statusbar_get_context_id (    statusbar,
                                                        "installed status");
    data.status_id = 0;
    data.statusbar = statusbar;
    
    /* save treeview to data object */
    data.tree = GTK_TREE_VIEW(treeview);
    
    /* entry and button callback */
    entry = GTK_WIDGET ( gtk_builder_get_object (builder, "entry") );
    g_signal_connect(   G_OBJECT(entry), "activate",
                        G_CALLBACK (search_click), 
                        &data);
                        
    button = GTK_WIDGET ( gtk_builder_get_object (builder, "button") );
    g_signal_connect(   G_OBJECT(button), "clicked",
                        G_CALLBACK (search_click), 
                        &data);    
                                            
    data.entry = GTK_ENTRY(entry);
    
    /* setup GtkTextView */
    text_all = GTK_WIDGET ( gtk_builder_get_object (builder, "text_all") );
    data.text_all = GTK_TEXT_VIEW( text_all );
    text_description = GTK_WIDGET ( gtk_builder_get_object (builder, "text_description") );
    data.text_description = GTK_TEXT_VIEW ( text_description );
    text_special = GTK_WIDGET ( gtk_builder_get_object (builder, "text_special") );
    data.text_special = GTK_TEXT_VIEW ( text_special );
    
    /* setup all_tab label */
    text_all_label = GTK_WIDGET ( gtk_builder_get_object (builder, "text_all_label") );
    data.text_all_label = GTK_LABEL( text_all_label );
    
    /* setup description_tab label */
    text_description_label = GTK_WIDGET ( gtk_builder_get_object (builder, "text_description_label") );
    data.text_description_label = GTK_LABEL( text_description_label );
    
    /* setup special_tab label */
    text_special_label = GTK_WIDGET ( gtk_builder_get_object (builder, "text_special_label") );
    data.text_special_label = GTK_LABEL( text_special_label );
    
    /* change window title */
    gtk_window_set_title(   GTK_WINDOW(data.gladewindow), 
                            "gaptsearch");
    
    /* show widget, unref builder, start main loop */
    g_object_unref (G_OBJECT (builder));
    gtk_widget_show (data.gladewindow);
    gtk_main ();
   
    return 0;
}
